@@ -3,7 +3,7 @@ package HTML::Mason;
 # This program is free software; you can redistribute it and/or modify it
 # under the same terms as Perl itself.
 
-$HTML::Mason::VERSION = '0.5';
+$HTML::Mason::VERSION = '0.6';
 
 use HTML::Mason::Parser;
 use HTML::Mason::Interp;
@@ -15,8 +15,5 @@ if ($ENV{MOD_PERL} || $HTML::Mason::IN_DEBUG_FILE) {
 	die "HTML::Mason::ApacheHandler class is not defined.  If you are in a mod_perl environment, this should have been brought in automatically.  Send a bug report to the Mason development team, and for now, try using HTML::Mason::ApacheHandler manually in your handler.pl.\n";
     }
 }
-*HTML::Mason::Preview::handle_preview_request = sub {
-    die "HTML::Mason::Preview class is not defined.  With Mason 0.5 and greater you need to manually \"use HTML::Mason::Preview\" in your handler.pl.\n";
-};
 
 1;
