@@ -13,28 +13,31 @@ use strict;
 use Carp;
 use File::Path;
 use File::Basename;
-use File::Recurse;
+use File::Find;
 use IO::File;
 use IO::Seekable;
 use HTML::Mason::Parser;
 use HTML::Mason::Tools qw(read_file pkg_loaded);
+use HTML::Mason::Commands qw();
 use HTML::Mason::Config;
+require Time::HiRes if $HTML::Mason::Config{use_time_hires};
+
 use vars qw($AUTOLOAD $_SUB %_ARGS);
-my @used = ($HTML::Mason::Commands::INTERP);
+my @_used = ($HTML::Mason::CODEREF_NAME,$::opt_P);
 
 my %fields =
     (alternate_sources => undef,
+     allow_recursive_autohandlers => 0,
      comp_root => undef,
      code_cache_mode => 'all',
      current_time => 'real',
      data_dir => undef,
      system_log_file => undef,
-     system_log_events => '',
      system_log_separator => "\cA",
      max_recurse => 16,
      parser => undef,
      preloads => [],
-     static_file_root => '',
+     static_file_root => undef,
      use_data_cache => 1,
      use_object_files => 1,
      use_reload_file => 0,
@@ -43,7 +46,7 @@ my %fields =
      );
 # Minor speedup: create anon. subs to reduce AUTOLOAD calls
 foreach my $f (keys %fields) {
-    next if $f eq 'current_time';  # don't overwrite real sub.
+    next if $f =~ /^current_time|system_log_events$/;  # don't overwrite real sub.
     no strict 'refs';
     *{$f} = sub {my $s=shift; return @_ ? ($s->{$f}=shift) : $s->{$f}};
 }
@@ -57,28 +60,32 @@ sub new
 	data_cache_store => {},
         code_cache => {},
 	files_written => [],
-	hooks => [],
-	hook_index => {},
+	hooks => {},
 	last_reload_time => 0,
 	last_reload_file_pos => 0,
 	out_method => sub { print $_[0] },
+	system_log_fh => undef,
+	system_log_events_hash => undef
     };
     my (%options) = @_;
-    my ($rootDir,$outMethod);
+    my ($rootDir,$outMethod,$systemLogEvents);
     while (my ($key,$value) = each(%options)) {
 	if (exists($fields{$key})) {
 	    $self->{$key} = $value;
 	} elsif ($key eq 'out_method') {
 	    $outMethod = $value;
+	} elsif ($key eq 'system_log_events') {
+	    $systemLogEvents = $value;
 	} else {
 	    die "HTML::Mason::Interp::new: invalid option '$key'\n";
 	}
     }
-    $self->{data_cache_dir} ||= ($self->{data_dir} . "/cache");
     die "HTML::Mason::Interp::new: must specify value for comp_root\n" if !$self->{comp_root};
     die "HTML::Mason::Interp::new: must specify value for data_dir\n" if !$self->{data_dir};
+    $self->{data_cache_dir} ||= ($self->{data_dir} . "/cache");
     bless $self, $class;
     $self->out_method($outMethod) if ($outMethod);
+    $self->system_log_events($systemLogEvents) if ($systemLogEvents);
     $self->_initialize;
     return $self;
 }
@@ -106,21 +113,14 @@ sub _initialize
     }
     
     #
-    # Open system log file, and convert events string to hash
+    # Open system log file
     #
-    if ($self->system_log_events) {
+    if ($self->{system_log_events_hash}) {
 	$self->system_log_file($self->data_dir . "/etc/system.log") if !$self->system_log_file;
 	my $fh = new IO::File ">>".$self->system_log_file
 	    or die "Couldn't open system log file ".$self->{system_log_file}." for append";
 	$fh->autoflush(1);
-	my $eventstr = $self->system_log_events;
-	$eventstr =~ s/\s//g;
-	my %opts = map( ($_, 1), split /\|/, $eventstr);
-	@opts{qw(REQUEST CACHE COMP_LOAD)} = (1,1,1) if $opts{ALL};
-	@opts{qw(CACHE_READ CACHE_WRITE)} = (1,1) if $opts{CACHE};
-	@opts{qw(REQ_START REQ_END)} = (1,1) if $opts{REQUEST};
-	$self->system_log_file($fh);
-	$self->system_log_events(\%opts);
+	$self->{system_log_fh} = $fh;
     }
     
     #
@@ -134,13 +134,15 @@ sub _initialize
 	    my $fullPath = $self->comp_root . $p;
 	    $fullPath =~ s@/$@@g;
 	    if (-d $fullPath) {
-		recurse {
-		    if (-f $_) {
-			my $compPath = substr($_,$slen);
+		my $sub = sub {
+		    if (-f) {
+			my $file = $_;
+			$file =~ s/^\./$fullPath/;
+			my $compPath = substr($file,$slen);
 			$self->load($compPath);
-			return 0;
 		    }
-		} $fullPath;
+		};
+		find($sub,$fullPath);
 	    } elsif (-f $fullPath) {
 		my $compPath = substr($fullPath,$slen);
 		$self->load($compPath);
@@ -191,13 +193,17 @@ sub exec {
 
     $self->{stack} = [];
     $self->{exec_state} = {
-	suppressed_hooks => {},
 	abort_flag => 0,
 	abort_retval => undef,
+	autohandler_next => undef,
 	error_flag => 0,
 	request_code_cache => {}
     };
+    while (my ($type,$href) = each(%{$self->{hooks}})) {
+	$self->{exec_state}->{"hooks_$type"} = [values(%$href)] if (%$href);
+    }
     $self->check_reload_file if ($self->use_reload_file);
+
     return $self->exec_next($callPath,%args);
 }
 
@@ -229,57 +235,48 @@ sub check_reload_file {
 
 sub exec_next {
     my ($self, $callPath, %args) = @_;
-    my ($err);
+    my (@info);
 
-    $callPath =~ s@/$@@ if ($callPath ne '/');
-    my $allowHandlers = $args{ALLOW_HANDLERS};
+    # Process and remove special arguments.
     my $store = $args{STORE};
-    delete($args{ALLOW_HANDLERS});
     delete($args{STORE});
 
+    # Check for autohandler.
+    if (!@{$self->{stack}}) {
+	(my $callDir = $callPath) =~ s|/[^/]*$||;
+	if (!$self->{allow_recursive_autohandlers}) {
+	    @info = $self->load("$callDir/autohandler");
+	} else {
+	    while ($callDir && !(@info = $self->load("$callDir/autohandler"))) {
+		my ($basename,$dirname) = fileparse($callDir);
+		$dirname =~ s/^\.//;    # certain versions leave ./ in $dirname
+		$callDir = substr($dirname,0,-1);
+	    }
+	}
+	if (@info) {
+	    $self->{exec_state}->{autohandler_next} = [$callPath,\%args];
+	    $callPath = "$callDir/autohandler";
+	}
+    }
+					   
     #
     # Load the component into a subroutine. If code caching is
     # per-request, then check and modify the per-request cache as
     # necessary.
     #
-    my (@info);
-    if ($self->{code_cache_mode} eq 'request') {
-	if (my $inforef = $self->{exec_state}->{request_code_cache}->{$callPath}) {
-	    @info = @$inforef;
+    if (!@info) {
+	if ($self->{code_cache_mode} eq 'request') {
+	    if (my $inforef = $self->{exec_state}->{request_code_cache}->{$callPath}) {
+		@info = @$inforef;
+	    } else {
+		(@info) = $self->load($callPath);
+		$self->{exec_state}->{request_code_cache}->{$callPath} = [@info];
+	    }
 	} else {
 	    (@info) = $self->load($callPath);
-	    $self->{exec_state}->{request_code_cache}->{$callPath} = [@info];
-	}
-    } else {
-	(@info) = $self->load($callPath);
-    }
-    if (!@info) {
-	if (!$allowHandlers) {
-	    die "could not find component for path '$callPath'\n";
-	} else {
-	    #
-	    # This hack implements the ALLOW_HANDLERS flag for
-	    # backward compatibility with Scribe.  Looks for home and
-	    # dhandler files when component not found.  Hopefully can
-	    # remove someday soon.
-	    #
-	    my $pathInfo = '';
-	    my $p = $callPath;
-	    if (!(@info = $self->load("$p/home"))) {
-		while (!(@info = $self->load("$p/dhandler"))) {
-		    my ($basename,$dirname) = fileparse($p);
-		    $pathInfo = "/$basename$pathInfo";
-		    $p = substr($dirname,0,-1);
-		    die "could not find component for path '$callPath'\n" if $p !~ /\S/;
-		}
-		$callPath = "$p/dhandler";
-		my $r = $self->vars('server') or die "No ALLOW_HANDLERS outside of Apache";
-		$r->path_info($pathInfo);
-	    } else {
-		$callPath = "$p/home";
-	    }
 	}
     }
+    die "could not find component for path '$callPath'\n" if (!@info);
     my ($sub,$sourceFile) = @info;
 
     #
@@ -288,10 +285,12 @@ sub exec_next {
     unshift(@{$self->{stack}},{});
     
     #
-    # $INTERP is a dynamically scoped global containing this
-    # interpreter.
+    # $INTERP is a global containing this interpreter. This needs to
+    # be defined in the HTML::Mason::Commands package, as well
+    # as the component package if that is different.
     #
     local $HTML::Mason::Commands::INTERP = $self;
+    $self->set_global(INTERP=>$self) if ($self->{parser}->{in_package} ne 'HTML::Mason::Commands');
 
     #
     # Check for maximum recursion.
@@ -312,8 +311,6 @@ sub exec_next {
     $locals->{sourceFile} = $sourceFile;
     ($locals->{parentPath}) = ($callPath =~ m@^(.*)/([^/]*)$@);
     $locals->{parentPath} = '/' if ($locals->{parentPath} !~ /\S/);
-    $locals->{callFunc} = $sub;
-    $locals->{callArgs} = \%args;
     if ($store) {
 	die "exec_next: STORE is not a scalar reference" if ref($store) ne 'SCALAR';
 	$$store = '';
@@ -323,12 +320,14 @@ sub exec_next {
     } else {
 	$locals->{sink} = $self->{stack}->[1]->{sink};
     }
+    $locals->{callFunc} = $sub;
+    $locals->{callArgs} = \%args;
     $self->{stack}->[0] = $locals;
 
     #
     # Call start_comp hooks.
     #
-    $self->call_hooks(type=>'start_comp');
+    $self->call_hooks('start_comp');
 
     #
     # CODEREF_NAME maps component coderefs to component names (for profiling)
@@ -353,7 +352,7 @@ sub exec_next {
     #
     # If an error occurred...
     #
-    $err = $@;
+    my $err = $@;
     if ($err) {
 	
 	#
@@ -395,7 +394,7 @@ sub exec_next {
     #
     # Call end_comp hooks.
     #
-    $self->call_hooks(type=>'end_comp');
+    $self->call_hooks('end_comp');
 
     #
     # Pop stack and return.
@@ -404,6 +403,17 @@ sub exec_next {
     return wantarray ? @result : $result;
 }
 
+#
+# Abort out of current execution.
+#
+sub abort
+{
+    my ($self) = @_;
+    $self->{exec_state}->{abort_flag} = 1;
+    $self->{exec_state}->{abort_retval} = $_[1];
+    die "aborted";
+}
+
 sub debug_hook
 {
     1;
@@ -415,11 +425,11 @@ sub debug_hook
 sub pure_text_handler
 {
     my $interp = $HTML::Mason::Commands::INTERP;
-    $interp->call_hooks(type=>'start_primary');
+    $interp->call_hooks('start_primary');
     my $srcfile = $interp->locals->{sourceFile};
     my $srctext = read_file($srcfile);
     $interp->locals->{sink}->($srctext);
-    $interp->call_hooks(type=>'end_primary');
+    $interp->call_hooks('end_primary');
     return undef;
 }
 
@@ -431,55 +441,73 @@ sub pure_text_handler
 #
 sub load {
     my ($self,$path) = @_;
-    my ($sub,$err,$maxfilemod,$srcfile,$objfile,$objfilemod,$srcfilemod);
+    my ($sub,$err,$maxfilemod,$srcfile,$objfile,$objfilemod,$srcfilemod) =
+       (undef);  # kludge to load $sub, prevent "use of uninit .." error
+    my (@srcstat, @objstat, $objisfile);
     my $codeCache = $self->{code_cache};
     my $compRoot = $self->comp_root;
+    $srcfile = $compRoot . $path;
 
     #
-    # If using reload file, check cache first.
+    # If using reload file, assume that we have a cached subroutine or
+    # object file (also assume we're using obj files). If no obj file
+    # exists, this is likely a dhandler request.
     #
-    if (exists($codeCache->{$path}) && $self->use_reload_file) {
-	return @{$codeCache->{$path}->{info}};
+    if ($self->use_reload_file) {
+	return @{$codeCache->{$path}->{info}} if exists($codeCache->{$path});
+
+	$objfile = $self->object_dir . substr($srcfile,length($compRoot));
+	return () unless (-f $objfile);   # component not found
+	
+	$self->write_system_log('COMP_LOAD', $path);	# log the load event
+	$self->parser->evaluate(script_file=>$objfile,code=>\$sub,error=>\$err);
+	if ($err) {
+	    $sub = sub { die "Error while loading '$objfile' at runtime:\n$err\n" };
+	} elsif (!$sub) {
+	    $sub = \&pure_text_handler;
+	}
+	
+	my @info = ($sub,$srcfile);
+	if ($self->{code_cache_mode} eq 'all') {
+	    $codeCache->{$path}->{info} = \@info;
+	}
+	return @info;
     }
     
     #
     # Determine source and (possibly) object filename.
     # If alternate sources are defined, check those first.
     #
-    $srcfile = $compRoot . $path;
     if (defined($self->alternate_sources)) {
 	my @alts = $self->alternate_sources->('comp',$path);
 	foreach my $alt (@alts) {
-	    if (-f $compRoot . $alt) {
+	    @srcstat = stat ($compRoot . $alt);
+	    if (-f _) {
 		$srcfile = $compRoot . $alt;
 		last;
 	    }
 	}
     }
-    return () if (!-f $srcfile);
+    @srcstat = stat $srcfile unless @srcstat;
+    return () unless (-f _);
+    $srcfilemod = $srcstat[9];
+    
     if ($self->use_object_files) {
 	$objfile = $self->object_dir . substr($srcfile,length($compRoot));
+	@objstat = stat $objfile;
+	$objisfile = -f _;
     }
     
-    #
-    # If not using reload file, check the last modified times of
-    # source and object filename.
-    #
-    if (!$self->use_reload_file) {
-	$srcfilemod = [stat($srcfile)]->[9];
-    }
-
     #
     # If code cache contains an up to date entry for this path,
     # use the cached sub.
     #
-    if (exists($codeCache->{$path}) && ($self->use_reload_file || $codeCache->{$path}->{lastmod} >= $srcfilemod) && $codeCache->{$path}->{info}->[1] eq $srcfile) {
+    if (exists($codeCache->{$path})                    and
+	$codeCache->{$path}->{lastmod} >= $srcfilemod  and
+	$codeCache->{$path}->{info}->[1] eq $srcfile) {
 	return @{$codeCache->{$path}->{info}};
     } else {
-	if ($self->use_reload_file) {
-	    $srcfilemod = [stat($srcfile)]->[9];
-	}
-	$objfilemod = (defined($objfile) && (-f $objfile)) ? [stat($objfile)]->[9] : 0;
+	$objfilemod = (defined($objfile) and $objisfile) ? $objstat[9] : 0;
 	
 	#
 	# Determine a subroutine.
@@ -490,23 +518,23 @@ sub load {
 	#
 	$self->write_system_log('COMP_LOAD', $path);	# log the load event
 
-	if ($objfile && $objfilemod >= $srcfilemod) {
-	    if ([stat($objfile)]->[7]==0) {
+	if ($objfile and $objfilemod >= $srcfilemod) {
+	    $self->parser->evaluate(script_file=>$objfile,code=>\$sub,error=>\$err);
+	    if ($err) {
+		$sub = sub { die "Error while loading '$objfile' at runtime:\n$err\n" };
+	    } elsif (!$sub) {
 		$sub = \&pure_text_handler;
-	    } else {
-		$self->parser->evaluate(script_file=>$objfile,code=>\$sub,error=>\$err);
-		if ($err) {
-		    $sub = sub { die "Error while loading '$objfile' at runtime:\n$err\n" };
-		}
 	    }
-	}
-
 	#
 	# Parse the source file, and possibly save result in object file.
 	#
-	else {
+	} else {
 	    my ($script,$objtext,$errmsg,$pureTextFlag);
-	    my $success = $self->parser->parse(script_file=>$srcfile, result_text=>\$objtext, error=>\$errmsg, result_code=>\$sub, pure_text_flag=>\$pureTextFlag);
+	    my $success = $self->parser->parse(script_file=>$srcfile,
+					       result_text=>\$objtext,
+					       error=>\$errmsg,
+					       result_code=>\$sub,
+					       pure_text_flag=>\$pureTextFlag);
 	    $sub = \&pure_text_handler if ($pureTextFlag);
 	    if (!$success) {
 		$errmsg = "Error during compilation of $srcfile:\n$errmsg\n";
@@ -519,7 +547,7 @@ sub load {
 		$errmsg =~ s/\'/\\\'/g;
 		$errmsg =~ s/\(eval [0-9]\) //g;
 		($errmsg) = ($errmsg =~ /^(.*)$/s);
-		my $script = "return sub { die '$errmsg' };";
+		my $script = "sub { die '$errmsg' };";
 		$sub = eval($script);
 	    }
 	    if ($objfile) {
@@ -546,7 +574,7 @@ sub load {
 	#
 	my @info = ($sub,$srcfile);
 	if ($self->{code_cache_mode} eq 'all') {
-	    $codeCache->{$path}->{lastmod} = $srcfilemod if !$self->use_reload_file;
+	    $codeCache->{$path}->{lastmod} = $srcfilemod;
 	    $codeCache->{$path}->{info} = \@info;
 	}
 	return @info;
@@ -604,6 +632,53 @@ sub vars
     }
 }
 
+sub set_global
+{
+    my ($self, $decl, @values) = @_;
+    my ($prefix, $name) = ($decl =~ /^[\$@%]/) ? (substr($decl,0,1),substr($decl,1)) : ("\$",$decl);
+
+    my $varname = sprintf("%s::%s",$self->{parser}->{in_package},$name);
+    if ($prefix eq "\$") {
+	no strict 'refs'; $$varname = $values[0];
+    } elsif ($prefix eq "\@") {
+	no strict 'refs'; @$varname = @values;
+    } else {
+	no strict 'refs'; %$varname = @values;
+    }
+}
+
+#
+# Allow scalar or hash reference as argument to system_log_events.
+#
+sub system_log_events
+{
+    my ($self, $value) = @_;
+    if (defined($value)) {
+	if (!ref($value)) {
+	    $value =~ s/\s//g;
+	    my %opts = map( ($_, 1), split /\|/, $value);
+	    @opts{qw(REQUEST CACHE COMP_LOAD)} = (1,1,1) if $opts{ALL};
+	    @opts{qw(CACHE_READ CACHE_WRITE)} = (1,1) if $opts{CACHE};
+	    @opts{qw(REQ_START REQ_END)} = (1,1) if $opts{REQUEST};
+	    $self->{system_log_events_hash} = \%opts;
+	} elsif (ref($value) eq 'HASH') {
+	    $self->{system_log_events_hash} = $value;
+	} else {
+	    die "system_log_events: argument must be a scalar or hash reference";
+	}
+    }
+    return $self->{system_log_events};
+}
+
+#
+# Determine if the specified event should be logged.
+#
+sub system_log_event_check
+{
+    my ($self,$flag) = @_;
+    return ($self->{system_log_fh} && $self->{system_log_events_hash}->{$flag});
+}
+
 #
 # Allow scalar or code reference as argument to out_method.
 #
@@ -612,7 +687,7 @@ sub out_method
     my ($self, $value) = @_;
     if (defined($value)) {
 	if (ref($value) eq 'SCALAR') {
-	    $self->{out_method} = sub { $$value .= $_[0] };
+	    $self->{out_method} = sub { $$value .= $_[0] if defined($_[0]) };
 	} elsif (ref($value) eq 'CODE') {
 	    $self->{out_method} = $value;
 	} else {
@@ -641,80 +716,41 @@ sub push_files_written
 #
 # Hook functions.
 #
+my @hookTypes = qw(start_comp start_primary end_primary end_comp start_file end_file);
+my %hookTypeMap = map(($_,1),@hookTypes);
+
 sub add_hook {
     my ($self, %args) = @_;
-    my @validTypes = qw(start_comp start_primary end_primary end_comp start_file end_file);
     foreach (qw(name type code)) {
 	die "add_hook: must specify $_\n" if !exists($args{$_});
     }
-    die "add_hook: type must be one of ".join(",",@validTypes)."\n" if !grep($_ eq $args{type},@validTypes);
+    die "add_hook: type must be one of ".join(",",@hookTypes)."\n" if !$hookTypeMap{$args{type}};
     die "add_hook: code must be a code reference\n" if ref($args{code}) ne 'CODE';
-    $args{order} = 50 if !exists($args{order});
-    push(@{$self->{hooks}},{%args});
-    $self->index_hooks();
+    $self->{hooks}->{$args{type}}->{$args{name}} = $args{code};
 }
 
-sub index_hooks {
-    my ($self) = @_;
-    my %hookindex;
-    foreach my $hook (@{$self->{hooks}}) {
-	my $name = $hook->{name};
-	my $type = $hook->{type};
-	push(@{$hookindex{"$name\cA$type"}},$hook);
-	push(@{$hookindex{"*\cA$type"}},$hook);
-	push(@{$hookindex{"$name\cA*"}},$hook);
-    }
-    my @keys = keys(%hookindex);
-    foreach my $key (@keys) {
-	my @hooklst = sort {$a->{order} <=> $b->{order}} @{$hookindex{$key}};
-	$hookindex{$key} = [@hooklst];
-    }
-    $self->{hook_index} = {%hookindex};
-}
-
-sub find_hooks {
+sub remove_hook {
     my ($self, %args) = @_;
-    my $name = $args{name} || '*';
-    my $type = $args{type} || '*';
-    my $lref = $self->{hook_index}->{"$name\cA$type"};
-    return $lref ? @$lref : ();
-}
-
-sub remove_hooks {
-    my ($self, %args) = @_;
-    my @matchhooks = $self->find_hooks(%args);
-    my %remhash;
-    foreach (@matchhooks) { $remhash{$_}=1 }
-    my @keephooks = grep(!$remhash{$_},@{$self->{hooks}});
-    $self->{hooks} = [@keephooks];
-    $self->index_hooks();
-}
-
-sub suppress_hooks {
-    my ($self, %args) = @_;
-    my @matchhooks = $self->find_hooks(%args);
-    foreach my $hook (@matchhooks) {
-	$self->{exec_state}->{suppressed_hooks}->{$hook} = 1;
+    foreach (qw(name type)) {
+	die "remove_hook: must specify $_\n" if !exists($args{$_});
     }
+    delete($self->{hooks}->{$args{type}}->{$args{name}});
 }
 
-sub unsuppress_hooks {
+sub suppress_hook {
     my ($self, %args) = @_;
-    my @matchhooks = $self->find_hooks(%args);
-    foreach my $hook (@matchhooks) {
-	delete($self->{exec_state}->{suppressed_hooks}->{$hook});
+    foreach (qw(name type)) {
+	die "suppress_hook: must specify $_\n" if !exists($args{$_});
     }
+    my $code = $self->{hooks}->{$args{type}}->{$args{name}};
+    $self->{exec_state}->{"hooks_$args{type}"} = [grep($_ ne $code,@{$self->{"hooks_$args{type}"}})];
 }
 
 sub call_hooks {
-    my ($self, %args) = @_;
-    my @matchhooks = $self->find_hooks(%args);
-    my @params;
-    @params = @{$args{params}} if (defined($args{params}));
-
-    foreach my $hook (@matchhooks) {
-	if (!$self->{exec_state}->{suppressed_hooks}->{$hook}) {
-	    $hook->{code}->($self,@params);
+    my ($self, $type, @params) = @_;
+    if ($self->{exec_state}->{"hooks_$type"}) {
+	foreach my $code (@{$self->{exec_state}->{"hooks_$type"}}) {
+	    $code->($self, @params);
 	}
     }
 }
@@ -724,14 +760,14 @@ sub call_hooks {
 # Each line begins with the time, pid, and event name.
 #
 # We only print the line if the log file handle is defined AND the
-# event name is in our system_log_events hash.
+# event name is in system_log_events_hash.
 #
 sub write_system_log {
     my $self = shift;
 
-    if (defined $self->system_log_file && $self->system_log_events->{$_[0]}) {
-	my $time = (pkg_loaded("Time::HiRes") ? scalar(Time::HiRes::gettimeofday()) : time);
-	$self->system_log_file->print(join ($self->system_log_separator,
+    if ($self->{system_log_fh} && $self->{system_log_events_hash}->{$_[0]}) {
+	my $time = ($HTML::Mason::Config{use_time_hires} ? scalar(Time::HiRes::gettimeofday()) : time);
+	$self->{system_log_fh}->print(join ($self->system_log_separator,
 					    $time,                  # current time
 					    $_[0],                  # event name
 					    $$,                     # pid
